@@ -10,7 +10,7 @@ import subprocess
 import tempfile
 import email
 import email.policy
-from email.utils import getaddresses, parseaddr, formataddr
+from email.utils import getaddresses, parseaddr
 from email.message import Message
 
 import urwid
@@ -24,10 +24,14 @@ from .globals import MoveCommand
 from .globals import CommandCanceled
 from .common import RetagPromptCommand
 from .envelope import SendCommand
-from ..completion import ContactsCompleter, PathCompleter
+from ..completion.contacts import ContactsCompleter
+from ..completion.path import PathCompleter
 from ..db.utils import decode_header
+from ..db.utils import formataddr
+from ..db.utils import get_body_part
 from ..db.utils import extract_headers
-from ..db.utils import extract_body
+from ..db.utils import clear_my_address
+from ..db.utils import ensure_unique_address
 from ..db.envelope import Envelope
 from ..db.attachment import Attachment
 from ..db.errors import DatabaseROError
@@ -135,20 +139,21 @@ class ReplyCommand(Command):
         mail = self.message.get_email()
 
         # set body text
-        name, address = self.message.get_author()
+        name, address = parseaddr(mail['From'])
         timestamp = self.message.get_date()
         qf = settings.get_hook('reply_prefix')
         if qf:
-            quotestring = qf(name, address, timestamp, ui=ui, dbm=ui.dbman)
+            quotestring = qf(name, address, timestamp,
+                             message=mail, ui=ui, dbm=ui.dbman)
         else:
             quotestring = 'Quoting %s (%s)\n' % (name or address, timestamp)
         mailcontent = quotestring
         quotehook = settings.get_hook('text_quote')
         if quotehook:
-            mailcontent += quotehook(self.message.accumulate_body())
+            mailcontent += quotehook(self.message.get_body_text())
         else:
             quote_prefix = settings.get('quote_prefix')
-            for line in self.message.accumulate_body().splitlines():
+            for line in self.message.get_body_text().splitlines():
                 mailcontent += quote_prefix + line + '\n'
 
         envelope = Envelope(bodytext=mailcontent, replied=self.message)
@@ -202,7 +207,7 @@ class ReplyCommand(Command):
             # make sure that our own address is not included
             # if the message was self-sent, then our address is not included
             MFT = mail.get_all('Mail-Followup-To', [])
-            followupto = self.clear_my_address(account, MFT)
+            followupto = clear_my_address(account, MFT)
             if followupto and settings.get('honor_followup_to'):
                 logging.debug('honor followup to: %s', ', '.join(followupto))
                 recipients = followupto
@@ -213,17 +218,16 @@ class ReplyCommand(Command):
 
                 # append To addresses if not replying to self sent message
                 if not account.matches_address(sender_address):
-                    cleared = self.clear_my_address(
-                        account, mail.get_all('To', []))
+                    cleared = clear_my_address(account,
+                                               mail.get_all('To', []))
                     recipients.extend(cleared)
 
                 # copy cc for group-replies
                 if 'Cc' in mail:
-                    cc = self.clear_my_address(
-                        account, mail.get_all('Cc', []))
+                    cc = clear_my_address(account, mail.get_all('Cc', []))
                     envelope.add('Cc', decode_header(', '.join(cc)))
 
-        to = ', '.join(self.ensure_unique_address(recipients))
+        to = ', '.join(ensure_unique_address(recipients))
         logging.debug('reply to: %s', to)
 
         if self.listreply:
@@ -278,36 +282,6 @@ class ReplyCommand(Command):
                                               spawn=self.force_spawn,
                                               encrypt=encrypt))
 
-    @staticmethod
-    def clear_my_address(my_account, value):
-        """return recipient header without the addresses in my_account
-
-        :param my_account: my account
-        :type my_account: :class:`Account`
-        :param value: a list of recipient or sender strings (with or without
-            real names as taken from email headers)
-        :type value: list(str)
-        :returns: a new, potentially shortend list
-        :rtype: list(str)
-        """
-        new_value = []
-        for name, address in getaddresses(value):
-            if not my_account.matches_address(address):
-                new_value.append(formataddr((name, str(address))))
-        return new_value
-
-    @staticmethod
-    def ensure_unique_address(recipients):
-        """
-        clean up a list of name,address pairs so that
-        no address appears multiple times.
-        """
-        res = dict()
-        for name, address in getaddresses(recipients):
-            res[address] = name
-        urecipients = [formataddr((n, str(a))) for a, n in res.items()]
-        return sorted(urecipients)
-
 
 @registerCommand(MODE, 'forward', arguments=[
     (['--attach'], {'action': 'store_true', 'help': 'attach original mail'}),
@@ -346,20 +320,21 @@ class ForwardCommand(Command):
             timestamp = self.message.get_date()
             qf = settings.get_hook('forward_prefix')
             if qf:
-                quote = qf(name, address, timestamp, ui=ui, dbm=ui.dbman)
+                quote = qf(name, address, timestamp,
+                           message=mail, ui=ui, dbm=ui.dbman)
             else:
                 quote = 'Forwarded message from %s (%s):\n' % (
                     name or address, timestamp)
             mailcontent = quote
             quotehook = settings.get_hook('text_quote')
             if quotehook:
-                mailcontent += quotehook(self.message.accumulate_body())
+                mailcontent += quotehook(self.message.get_body_text())
             else:
                 quote_prefix = settings.get('quote_prefix')
-                for line in self.message.accumulate_body().splitlines():
+                for line in self.message.get_body_text().splitlines():
                     mailcontent += quote_prefix + line + '\n'
 
-            envelope.body = mailcontent
+            envelope.body_txt = mailcontent
 
             for a in self.message.get_attachments():
                 envelope.attach(a)
@@ -489,7 +464,7 @@ class EditNewCommand(Command):
                                 'signed', 'encrypted', 'unread', 'attachment'})
         tags = list(tags)
         # set body text
-        mailcontent = self.message.accumulate_body()
+        mailcontent = self.message.get_body_text()
         envelope = Envelope(bodytext=mailcontent, tags=tags)
 
         # copy selected headers
@@ -531,24 +506,36 @@ class EditNewCommand(Command):
     MODE, 'indent', help='change message/reply indentation',
     arguments=[(['indent'], {'action': cargparse.ValidatedStoreAction,
                              'validator': cargparse.is_int_or_pm})])
+@registerCommand(
+    MODE, 'togglemimetree', help='disply mime tree of the message',
+    forced={'mimetree': 'toggle'},
+    arguments=[(['query'], {'help': 'query used to filter messages to affect',
+                            'nargs': '*'})])
+@registerCommand(
+    MODE, 'togglemimepart', help='switch between html and plain text message',
+    forced={'mimepart': 'toggle'},
+    arguments=[(['query'], {'help': 'query used to filter messages to affect',
+                            'nargs': '*'})])
 class ChangeDisplaymodeCommand(Command):
 
     """fold or unfold messages"""
     repeatable = True
 
     def __init__(self, query=None, visible=None, raw=None, all_headers=None,
-                 indent=None, **kwargs):
+                 indent=None, mimetree=None, mimepart=False, **kwargs):
         """
         :param query: notmuch query string used to filter messages to affect
         :type query: str
         :param visible: unfold if `True`, fold if `False`, ignore if `None`
         :type visible: True, False, 'toggle' or None
-        :param raw: display raw message text.
+        :param raw: display raw message text
         :type raw: True, False, 'toggle' or None
         :param all_headers: show all headers (only visible if not in raw mode)
         :type all_headers: True, False, 'toggle' or None
         :param indent: message/reply indentation
         :type indent: '+', '-', or int
+        :param mimetree: show the mime tree of the message
+        :type mimetree: True, False, 'toggle' or None
         """
         self.query = None
         if query:
@@ -557,6 +544,8 @@ class ChangeDisplaymodeCommand(Command):
         self.raw = raw
         self.all_headers = all_headers
         self.indent = indent
+        self.mimetree = mimetree
+        self.mimepart = mimepart
         Command.__init__(self, **kwargs)
 
     def apply(self, ui):
@@ -602,6 +591,20 @@ class ChangeDisplaymodeCommand(Command):
             raw = not mt.display_source if self.raw == 'toggle' else self.raw
             all_headers = not mt.display_all_headers \
                 if self.all_headers == 'toggle' else self.all_headers
+            if self.mimepart:
+                if self.mimepart == 'toggle':
+                    message = mt.get_message()
+                    mimetype = {'plain': 'html', 'html': 'plain'}[
+                        message.mime_part.get_content_subtype()]
+                    mimepart = get_body_part(message.get_email(), mimetype)
+                elif self.mimepart is True:
+                    mimepart = ui.get_deep_focus().mimepart
+                mt.set_mimepart(mimepart)
+                ui.update()
+            if self.mimetree == 'toggle':
+                tbuffer.focus_selected_message()
+            mimetree = not mt.display_mimetree \
+                if self.mimetree == 'toggle' else self.mimetree
 
             # collapse/expand depending on new 'visible' value
             if visible is False:
@@ -614,6 +617,8 @@ class ChangeDisplaymodeCommand(Command):
                 mt.display_source = raw
             if all_headers is not None:
                 mt.display_all_headers = all_headers
+            if mimetree is not None:
+                mt.display_mimetree = mimetree
             mt.debug()
             # let the messagetree reassemble itself
             mt.reassemble()
@@ -636,8 +641,6 @@ class ChangeDisplaymodeCommand(Command):
                    'help': 'let the shell interpret the command'}),
     (['--notify_stdout'], {'action': 'store_true',
                            'help': 'display cmd\'s stdout as notification'}),
-    (['--field_key'], {'help': 'mailcap field key for decoding',
-                       'default': 'copiousoutput'}),
 ])
 class PipeCommand(Command):
 
@@ -647,8 +650,7 @@ class PipeCommand(Command):
     def __init__(self, cmd, all=False, separately=False, background=False,
                  shell=False, notify_stdout=False, format='raw',
                  add_tags=False, noop_msg='no command specified',
-                 confirm_msg='', done_msg=None, field_key='copiousoutput',
-                 **kwargs):
+                 confirm_msg='', done_msg=None, **kwargs):
         """
         :param cmd: shellcommand to open
         :type cmd: str or list of str
@@ -658,15 +660,15 @@ class PipeCommand(Command):
         :type separately: bool
         :param background: do not suspend the interface
         :type background: bool
-        :param notify_stdout: display command\'s stdout as notification message
-        :type notify_stdout: bool
         :param shell: let the shell interpret the command
         :type shell: bool
-
-                       'raw': message content as is,
-                       'decoded': message content, decoded quoted printable,
-                       'id': message ids, separated by newlines,
-                       'filepath': paths to message files on disk
+        :param notify_stdout: display command\'s stdout as notification message
+        :type notify_stdout: bool
+        :param format: what to pipe to the processes stdin. one of:
+            'raw': message content as is,
+            'decoded': message content, decoded quoted printable,
+            'id': message ids, separated by newlines,
+            'filepath': paths to message files on disk
         :type format: str
         :param add_tags: add 'Tags' header to the message
         :type add_tags: bool
@@ -677,8 +679,6 @@ class PipeCommand(Command):
         :type confirm_msg: str
         :param done_msg: notification message to show upon success
         :type done_msg: str
-        :param field_key: malcap field key for decoding
-        :type field_key: str
         """
         Command.__init__(self, **kwargs)
         if isinstance(cmd, str):
@@ -694,7 +694,6 @@ class PipeCommand(Command):
         self.noop_msg = noop_msg
         self.confirm_msg = confirm_msg
         self.done_msg = done_msg
-        self.field_key = field_key
 
     async def apply(self, ui):
         # abort if command unset
@@ -738,7 +737,7 @@ class PipeCommand(Command):
                     pipestrings.append(mail.as_string())
                 elif self.output_format == 'decoded':
                     headertext = extract_headers(mail)
-                    bodytext = extract_body(mail, field_key=self.field_key)
+                    bodytext = msg.get_body_text()
                     msgtext = '%s\n\n%s' % (headertext, bodytext)
                     pipestrings.append(msgtext)
 
@@ -1017,6 +1016,7 @@ class MoveFocusCommand(MoveCommand):
 
     def apply(self, ui):
         logging.debug(self.movement)
+        original_focus = ui.get_deep_focus()
         tbuffer = ui.current_buffer
         if self.movement == 'parent':
             tbuffer.focus_parent()
@@ -1047,18 +1047,28 @@ class MoveFocusCommand(MoveCommand):
         # TODO add 'next matching' if threadbuffer stores the original query
         # TODO: add next by date..
 
+        if original_focus != ui.get_deep_focus():
+            ui.update()
+
 
 @registerCommand(MODE, 'select')
 class ThreadSelectCommand(Command):
 
     """select focussed element:
         - if it is a message summary, toggle visibility of the message;
-        - if it is an attachment line, open the attachment"""
+        - if it is an attachment line, open the attachment
+        - if it is a mimepart, toggle visibility of the mimepart"""
     async def apply(self, ui):
         focus = ui.get_deep_focus()
         if isinstance(focus, AttachmentWidget):
             logging.info('open attachment')
             await ui.apply_command(OpenAttachmentCommand(focus.get_attachment()))
+        elif getattr(focus, 'mimepart', False):
+            if isinstance(focus.mimepart, Attachment):
+                await ui.apply_command(OpenAttachmentCommand(focus.mimepart))
+            else:
+                await ui.apply_command(ChangeDisplaymodeCommand(
+                    mimepart=True, mimetree='toggle'))
         else:
             await ui.apply_command(ChangeDisplaymodeCommand(visible='toggle'))
 
@@ -1111,7 +1121,7 @@ class TagCommand(Command):
     """manipulate message tags"""
     repeatable = True
 
-    def __init__(self, tags=u'', action='add', all=False, flush=True,
+    def __init__(self, tags='', action='add', all=False, flush=True,
                  **kwargs):
         """
         :param tags: comma separated list of tagstrings to set
