@@ -6,9 +6,9 @@
 import os
 import email
 import email.charset as charset
-from email.iterators import typed_subpart_iterator
 import email.policy
 import email.utils
+from email.errors import MessageError
 import tempfile
 import re
 import logging
@@ -51,7 +51,7 @@ def add_signature_headers(mail, sigs, error_msg):
     assert error_msg is None or isinstance(error_msg, str)
 
     if not sigs:
-        error_msg = error_msg or u'no signature found'
+        error_msg = error_msg or 'no signature found'
     elif not error_msg:
         try:
             key = crypto.get_key(sigs[0].fpr)
@@ -95,7 +95,7 @@ def get_params(mail, failobj=None, header='content-type', unquote=True):
     return {k.lower(): v for k, v in mail.get_params(failobj, header, unquote)}
 
 
-def _handle_signatures(original, message, params):
+def _handle_signatures(original_bytes, original, message, params):
     """Shared code for handling message signatures.
 
     RFC 3156 is quite strict:
@@ -103,6 +103,11 @@ def _handle_signatures(original, message, params):
     * the second is of type 'application/pgp-signature'
     * the second contains the detached signature
 
+    :param original_bytes: the original top-level mail raw bytes,
+        containing the segments against which signatures will be verified.
+        Necessary because parsing and re-serialising a Message isn't
+        byte-perfect, which interferes with signature validation.
+    :type original_bytes: bytes
     :param original: The original top-level mail. This is required to attache
         special headers to
     :type original: :class:`email.message.Message`
@@ -111,33 +116,58 @@ def _handle_signatures(original, message, params):
     :param params: the message parameters as returned by :func:`get_params`
     :type params: dict[str, str]
     """
-    malformed = None
-    if len(message.get_payload()) != 2:
-        malformed = u'expected exactly two messages, got {0}'.format(
-            len(message.get_payload()))
-    else:
-        ct = message.get_payload(1).get_content_type()
+    try:
+        nb_parts = len(message.get_payload()) if message.is_multipart() else 1
+        if nb_parts != 2:
+            raise MessageError(
+                f'expected exactly two messages, got {nb_parts}')
+
+        signature_part = message.get_payload(1)
+        ct = signature_part.get_content_type()
         if ct != _APP_PGP_SIG:
-            malformed = u'expected Content-Type: {0}, got: {1}'.format(
-                _APP_PGP_SIG, ct)
+            raise MessageError(
+               f'expected Content-Type: {_APP_PGP_SIG}, got: {ct}')
 
-    # TODO: RFC 3156 says the alg has to be lower case, but I've seen a message
-    # with 'PGP-'. maybe we should be more permissive here, or maybe not, this
-    # is crypto stuff...
-    if not params.get('micalg', 'nothing').startswith('pgp-'):
-        malformed = u'expected micalg=pgp-..., got: {0}'.format(
-            params.get('micalg', 'nothing'))
+        # TODO[dcbaker]: RFC 3156 says the alg has to be lower case, but I've
+        #  seen a message with 'PGP-'. maybe we should be more permissive here,
+        #  or maybe not, this is crypto stuff...
+        mic_alg = params.get('micalg', 'nothing')
+        if not mic_alg.startswith('pgp-'):
+            raise MessageError(f'expected micalg=pgp-..., got: {mic_alg}')
 
-    sigs = []
-    if not malformed:
-        try:
-            sigs = crypto.verify_detached(
-                message.get_payload(0).as_bytes(policy=email.policy.SMTP),
-                message.get_payload(1).get_payload(decode=True))
-        except GPGProblem as e:
-            malformed = str(e)
+        # RFC 3156 section 5 says that "signed message and transmitted message
+        # MUST be identical". We therefore need to validate against the message
+        # as it was originally sent.
 
-    add_signature_headers(original, sigs, malformed)
+        # The transmitted content and therefore the signed content are using
+        # CRLF as line delimiter, but our eml file has most likely been
+        # converted to UNIX LF line ending in the local storage.
+        if b'\r\n' not in original_bytes:
+            original_bytes = original_bytes.replace(b'\n', b'\r\n')
+
+        # The sender's signed canonical form often differs from the one
+        # produced by Python's standard lib (in the number of blank lines
+        # between multipart segments...). We therefore need to extract the
+        # signed part directly from the original byte string.
+        signed_boundary = b'\r\n--' + message.get_boundary().encode()
+        original_chunks = original_bytes.split(signed_boundary)
+        nb_chunks = len(original_chunks)
+        if nb_chunks != 4:
+            raise MessageError(
+                f'unexpected number of multipart chunks, got {nb_chunks}')
+
+        signed_chunk = original_chunks[1]
+        if len(signed_chunk) < len(b'\r\n'):
+            raise MessageError('signed chunk has an invalid length')
+
+        sigs = crypto.verify_detached(
+            signed_chunk[len(b'\r\n'):],
+            signature_part.get_payload(decode=True))
+
+        add_signature_headers(original, sigs, None)
+
+    except (GPGProblem, MessageError) as error:
+        add_signature_headers(original, [], str(error))
 
 
 def _handle_encrypted(original, message, session_keys=None):
@@ -162,13 +192,13 @@ def _handle_encrypted(original, message, session_keys=None):
 
     ct = message.get_payload(0).get_content_type()
     if ct != _APP_PGP_ENC:
-        malformed = u'expected Content-Type: {0}, got: {1}'.format(
+        malformed = 'expected Content-Type: {0}, got: {1}'.format(
             _APP_PGP_ENC, ct)
 
     want = 'application/octet-stream'
     ct = message.get_payload(1).get_content_type()
     if ct != want:
-        malformed = u'expected Content-Type: {0}, got: {1}'.format(want, ct)
+        malformed = 'expected Content-Type: {0}, got: {1}'.format(want, ct)
 
     if not malformed:
         # This should be safe because PGP uses US-ASCII characters only
@@ -210,33 +240,24 @@ def _handle_encrypted(original, message, session_keys=None):
                 add_signature_headers(original, sigs, '')
 
     if malformed:
-        msg = u'Malformed OpenPGP message: {0}'.format(malformed)
-        content = email.message_from_string(msg, policy=email.policy.SMTP)
+        msg = 'Malformed OpenPGP message: {0}'.format(malformed)
+        content = email.message_from_string(msg,
+                                            _class=email.message.EmailMessage,
+                                            policy=email.policy.SMTP)
         content.set_charset('utf-8')
         original.attach(content)
 
 
-def decrypted_message_from_file(handle, session_keys=None):
-    '''Reads a mail from the given file-like object and returns an email
-    object, very much like email.message_from_file. In addition to
-    that OpenPGP encrypted data is detected and decrypted. If this
-    succeeds, any mime messages found in the recovered plaintext
-    message are added to the returned message object.
-
-    :param handle: a file-like object
-    :param session_keys: a list OpenPGP session keys
-    :returns: :class:`email.message.Message` possibly augmented with
-              decrypted data
-    '''
-    return decrypted_message_from_message(email.message_from_file(handle),
-                                          session_keys)
-
-
-def decrypted_message_from_message(m, session_keys=None):
+def _decrypted_message_from_message(original_bytes, m, session_keys=None):
     '''Detect and decrypt OpenPGP encrypted data in an email object. If this
     succeeds, any mime messages found in the recovered plaintext
     message are added to the returned message object.
 
+    :param original_bytes: the original top-level mail raw bytes,
+        containing the segments against which signatures will be verified.
+        Necessary because parsing and re-serialising a Message isn't
+        byte-perfect, which interferes with signature validation.
+    :type original_bytes: bytes
     :param m: an email object
     :param session_keys: a list OpenPGP session keys
     :returns: :class:`email.message.Message` possibly augmented with
@@ -252,7 +273,7 @@ def decrypted_message_from_message(m, session_keys=None):
         # handle OpenPGP signed data
         if (m.get_content_subtype() == 'signed' and
                 p.get('protocol') == _APP_PGP_SIG):
-            _handle_signatures(m, m, p)
+            _handle_signatures(original_bytes, m, m, p)
 
         # handle OpenPGP encrypted data
         elif (m.get_content_subtype() == 'encrypted' and
@@ -270,7 +291,7 @@ def decrypted_message_from_message(m, session_keys=None):
 
                 if (sub.get_content_subtype() == 'signed' and
                         p.get('protocol') == _APP_PGP_SIG):
-                    _handle_signatures(m, sub, p)
+                    _handle_signatures(original_bytes, m, sub, p)
                 elif (sub.get_content_subtype() == 'encrypted' and
                       p.get('protocol') == _APP_PGP_ENC):
                     _handle_encrypted(m, sub, session_keys)
@@ -278,27 +299,17 @@ def decrypted_message_from_message(m, session_keys=None):
     return m
 
 
-def decrypted_message_from_string(s, session_keys=None):
-    '''Reads a mail from the given string. This is the equivalent of
-    :func:`email.message_from_string` which does nothing but to wrap
-    the given string in a StringIO object and to call
-    :func:`email.message_from_file`.
-
-    Please refer to the documentation of :func:`message_from_file` for
-    details.
-
-    '''
-    return decrypted_message_from_file(io.StringIO(s), session_keys)
-
-
 def decrypted_message_from_bytes(bytestring, session_keys=None):
     """Create a Message from bytes.
 
     :param bytes bytestring: an email message as raw bytes
     :param session_keys: a list OpenPGP session keys
     """
-    return decrypted_message_from_message(
-        email.message_from_bytes(bytestring, policy=email.policy.SMTP),
+    return _decrypted_message_from_message(
+        bytestring,
+        email.message_from_bytes(bytestring,
+                                 _class=email.message.EmailMessage,
+                                 policy=email.policy.SMTP),
         session_keys)
 
 
@@ -309,15 +320,15 @@ def extract_headers(mail, headers=None):
     one line "KEY: VALUE" for each requested header present in the mail.
 
     :param mail: the mail to use
-    :type mail: :class:`email.Message`
+    :type mail: :class:`email.message.EmailMessage`
     :param headers: headers to extract
     :type headers: list of str
     """
-    headertext = u''
+    headertext = ''
     if headers is None:
         headers = mail.keys()
     for key in headers:
-        value = u''
+        value = ''
         if key in mail:
             value = decode_header(mail.get(key, ''))
         headertext += '%s: %s\n' % (key, value)
@@ -389,7 +400,7 @@ def remove_cte(part, as_string=False):
     ..Note:: All this may be depricated in favour of
              `email.contentmanager.raw_data_manager` (v3.6+)
 
-    :param email.Message part: The part to decode
+    :param email.message.EmailMessage part: The part to decode
     :param bool as_string: If true return a str, otherwise return bytes
     :returns: The mail with any Content-Transfer-Encoding removed
     :rtype: Union[str, bytes]
@@ -446,70 +457,73 @@ def remove_cte(part, as_string=False):
             sp = helper.try_decode(bp)
         except UnicodeDecodeError as emsg:
             # the mail contains chars that are not enc-encoded.
-            # try again and just ignore those
+            # libmagic works better than just ignoring those
             logging.debug('Decoding failure: {}'.format(emsg))
-            sp = bp.decode(enc, errors='ignore')
+            sp = helper.try_decode(bp)
         return sp
     return bp
 
 
-def extract_body(mail, types=None, field_key='copiousoutput'):
-    """Returns a string view of a Message.
+MISSING_HTML_MSG = ("This message contains a text/html part that was not "
+                    "rendered due to a missing mailcap entry. "
+                    "Please refer to item 1 in our FAQ: "
+                    "http://alot.rtfd.io/en/latest/faq.html")
 
-    If the `types` argument is set then any encoding types there will be used
-    as the prefered encoding to extract. If `types` is None then
-    :ref:`prefer_plaintext <prefer-plaintext>` will be consulted; if it is True
-    then text/plain parts will be returned, if it is false then text/html will
-    be returned if present or text/plain if there are no text/html parts.
+
+def get_body_part(mail, mimetype=None):
+    """Returns an EmailMessage.
+
+    This consults :ref:`prefer_plaintext <prefer-plaintext>`
+    to determine if a "text/plain" alternative is preferred over a "text/html"
+    part.
 
     :param mail: the mail to use
-    :type mail: :class:`email.Message`
-    :param types: mime content types to use for body string
-    :type types: list[str]
+    :type mail: :class:`email.message.EmailMessage`
     :returns: The combined text of any parts to be used
     :rtype: str
     """
 
-    preferred = 'text/plain' if settings.get(
-        'prefer_plaintext') else 'text/html'
-    has_preferred = False
+    if not mimetype:
+        mimetype = 'plain' if settings.get('prefer_plaintext') else 'html'
+    preferencelist = {
+        'plain': ('plain', 'html'), 'html': ('html', 'plain')}[mimetype]
 
-    # see if the mail has our preferred type
-    if types is None:
-        has_preferred = list(typed_subpart_iterator(
-            mail, *preferred.split('/')))
+    body_part = mail.get_body(preferencelist)
+    if body_part is None:  # if no part matching preferredlist was found
+        return ""
 
-    body_parts = []
-    for part in mail.walk():
-        # skip non-leaf nodes in the mail tree
-        if part.is_multipart():
-            continue
+    return body_part
 
-        ctype = part.get_content_type()
 
-        if types is not None:
-            if ctype not in types:
-                continue
-        cd = part.get('Content-Disposition', '')
-        if cd.startswith('attachment'):
-            continue
-        # if the mail has our preferred type, we only keep this type
-        # note that if types != None, has_preferred always stays False
-        if has_preferred and ctype != preferred:
-            continue
+def extract_body_part(body_part):
+    """Returns a string view of a Message."""
+    displaystring = ""
+    rendered_payload = render_part(
+        body_part,
+        **{'field_key': 'view'} if body_part.get_content_type() == 'text/plain'
+        else {})
+    if rendered_payload:  # handler had output
+        displaystring = string_sanitize(rendered_payload)
+    elif body_part.get_content_type() == 'text/plain':
+        displaystring = string_sanitize(remove_cte(body_part, as_string=True))
+    else:
+        if body_part.get_content_type() == 'text/html':
+            displaystring = MISSING_HTML_MSG
+    return displaystring
 
-        if ctype == 'text/plain':
-            body_parts.append(string_sanitize(remove_cte(part, as_string=True)))
-        else:
-            rendered_payload = render_part(part)
-            if rendered_payload:  # handler had output
-                body_parts.append(string_sanitize(rendered_payload))
-            # mark as attachment
-            elif cd:
-                part.replace_header('Content-Disposition', 'attachment; ' + cd)
-            else:
-                part.add_header('Content-Disposition', 'attachment;')
-    return u'\n\n'.join(body_parts)
+
+def formataddr(pair):
+    """ this is the inverse of email.utils.parseaddr:
+    other than email.utils.formataddr, this
+    - *will not* re-encode unicode strings, and
+    - *will* re-introduce quotes around real names containing commas
+    """
+    name, address = pair
+    if not name:
+        return address
+    elif ',' in name:
+        name = "\"" + name + "\""
+    return "{0} <{1}>".format(name, address)
 
 
 def decode_header(header, normalize=False):
@@ -526,15 +540,9 @@ def decode_header(header, normalize=False):
     :type normalize: bool
     :rtype: str
     """
-    # some mailers send out incorrectly escaped headers
-    # and double quote the escaped realname part again. remove those
-    # RFC: 2047
-    regex = r'"(=\?.+?\?.+?\?[^ ?]+\?=)"'
-    value = re.sub(regex, r'\1', header)
-    logging.debug("unquoted header: |%s|", value)
+    logging.debug("unquoted header: |%s|", header)
 
-    # otherwise we interpret RFC2822 encoding escape sequences
-    valuelist = email.header.decode_header(value)
+    valuelist = email.header.decode_header(header)
     decoded_list = []
     for v, enc in valuelist:
         v = string_decode(v, enc)
@@ -553,3 +561,34 @@ def is_subdir_of(subpath, superpath):
     # return true, if the common prefix of both is equal to directory
     # e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
     return os.path.commonprefix([subpath, superpath]) == superpath
+
+
+def clear_my_address(my_account, value):
+    """return recipient header without the addresses in my_account
+
+    :param my_account: my account
+    :type my_account: :class:`Account`
+    :param value: a list of recipient or sender strings (with or without
+        real names as taken from email headers)
+    :type value: list(str)
+    :returns: a new, potentially shortend list
+    :rtype: list(str)
+    """
+    new_value = []
+    for name, address in email.utils.getaddresses(value):
+        if not my_account.matches_address(address):
+            new_value.append(formataddr((name, address)))
+    return new_value
+
+
+def ensure_unique_address(recipients):
+    """
+    clean up a list of name,address pairs so that
+    no address appears multiple times.
+    """
+    res = dict()
+    for name, address in email.utils.getaddresses(recipients):
+        res[address] = name
+    logging.debug(res)
+    urecipients = [formataddr((n, a)) for a, n in res.items()]
+    return sorted(urecipients)
